import typing

if typing.TYPE_CHECKING:
    from django.db.models.query import _QuerySet
    from django.utils.functional import _StrOrPromise as StrOrPromise
    from django.utils.functional import _StrPromise as StrPromise

    # Deprecated type aliases. Use the QuerySet class directly instead.
    QuerySetAny = _QuerySet
    ValuesQuerySet = _QuerySet
else:
    from django.db.models.query import QuerySet
    from django.utils.functional import Promise as StrPromise

    StrOrPromise = str | StrPromise
    # Deprecated type aliases. Use the QuerySet class directly instead.
    QuerySetAny = QuerySet
    ValuesQuerySet = QuerySet

__all__ = ["StrOrPromise", "StrPromise", "QuerySetAny", "ValuesQuerySet"]



import builtins
import logging
from collections.abc import Iterable
from typing import Any, Generic, TypeVar

from django import VERSION
from django.contrib.admin import ModelAdmin
from django.contrib.admin.options import BaseModelAdmin
from django.contrib.messages.views import SuccessMessageMixin
from django.contrib.sitemaps import Sitemap
from django.contrib.syndication.views import Feed
from django.core.files.utils import FileProxyMixin
from django.core.paginator import Paginator
from django.db.models.expressions import ExpressionWrapper
from django.db.models.fields import Field
from django.db.models.fields.related import ForeignKey
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)
from django.db.models.lookups import Lookup
from django.db.models.manager import BaseManager
from django.db.models.options import Options
from django.db.models.query import BaseIterable, ModelIterable, Prefetch, QuerySet, RawQuerySet
from django.forms.formsets import BaseFormSet
from django.forms.models import BaseModelForm, BaseModelFormSet, ModelChoiceField, ModelFormOptions
from django.utils.connection import BaseConnectionHandler, ConnectionProxy
from django.utils.functional import classproperty
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.edit import DeletionMixin, FormMixin
from django.views.generic.list import MultipleObjectMixin

__all__ = ["monkeypatch"]

logger = logging.getLogger(__name__)

_T = TypeVar("_T")
_VersionSpec = tuple[int, int]


class MPGeneric(Generic[_T]):
    """Create a data class to hold metadata about the generic classes needing monkeypatching.

    The `version` param is optional, and a value of `None` means that the monkeypatch is
    version-independent.

    This is slightly overkill for our purposes, but useful for future-proofing against any
    possible issues we may run into with this method.
    """

    def __init__(self, cls: type[_T], version: _VersionSpec | None = None) -> None:
        """Set the data fields, basic constructor."""
        self.version = version
        self.cls = cls

    def __repr__(self) -> str:
        """Better representation in tests and debug."""
        return "<MPGeneric: {}, versions={}>".format(self.cls, self.version or "all")


# certain django classes need to be generic, but lack the __class_getitem__ dunder needed to
# annotate them: https://github.com/typeddjango/django-stubs/issues/507
# this list stores them so `monkeypatch` can fix them when called
_need_generic: list[MPGeneric[Any]] = [
    MPGeneric(ModelAdmin),
    MPGeneric(SingleObjectMixin),
    MPGeneric(FormMixin),
    MPGeneric(DeletionMixin),
    MPGeneric(MultipleObjectMixin),
    MPGeneric(BaseModelAdmin),
    MPGeneric(Field),
    MPGeneric(Paginator),
    MPGeneric(BaseFormSet),
    MPGeneric(BaseModelForm),
    MPGeneric(BaseModelFormSet),
    MPGeneric(ModelChoiceField),
    MPGeneric(Feed),
    MPGeneric(Sitemap),
    MPGeneric(SuccessMessageMixin),
    MPGeneric(FileProxyMixin),
    MPGeneric(Lookup),
    MPGeneric(BaseConnectionHandler),
    MPGeneric(ExpressionWrapper),
    MPGeneric(ReverseManyToOneDescriptor),
    MPGeneric(ModelIterable),
    # These types do have native `__class_getitem__` method since django 3.1:
    MPGeneric(QuerySet, (3, 1)),
    MPGeneric(BaseManager, (3, 1)),
    # These types do have native `__class_getitem__` method since django 4.1:
    MPGeneric(ForeignKey, (4, 1)),
    MPGeneric(RawQuerySet),
    MPGeneric(classproperty),
    MPGeneric(ConnectionProxy),
    MPGeneric(ModelFormOptions),
    MPGeneric(Options),
    MPGeneric(BaseIterable),
    MPGeneric(ForwardManyToOneDescriptor),
    MPGeneric(ReverseOneToOneDescriptor),
    MPGeneric(Prefetch),
]


def monkeypatch(extra_classes: Iterable[type] | None = None, include_builtins: bool = True) -> None:
    """Monkey patch django as necessary to work properly with mypy."""
    # Add the __class_getitem__ dunder.
    suited_for_this_version = filter(
        lambda spec: spec.version is None or VERSION[:2] <= spec.version,
        _need_generic,
    )
    for el in suited_for_this_version:
        el.cls.__class_getitem__ = classmethod(lambda cls, *args, **kwargs: cls)
    if extra_classes:
        for cls in extra_classes:
            cls.__class_getitem__ = classmethod(lambda cls, *args, **kwargs: cls)  # type: ignore[attr-defined]

    # Add `reveal_type` and `reveal_locals` helpers if needed:
    if include_builtins:
        builtins.reveal_type = lambda obj, /: obj
        builtins.reveal_locals = lambda: None



from pathlib import Path
from typing import Any, TypedDict, type_check_only

from typing_extensions import NotRequired


@type_check_only
class TemplatesSetting(TypedDict):
    """Typing helper if you want to type `TEMPLATE` setting."""

    BACKEND: str
    NAME: NotRequired[str]
    DIRS: NotRequired[list[str | Path]]
    APP_DIRS: NotRequired[bool]
    OPTIONS: NotRequired[dict[str, Any]]


__all__ = ["TemplatesSetting"]



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any

    from django.db.models import Model

    class TypedDatabaseRouter:
        """
        Typed base class for Django's DATABASE_ROUTERS setting. At runtime this is just an alias to `object`.

        All methods are optional.

        Django documentation: https://docs.djangoproject.com/en/stable/topics/db/multi-db/#automatic-database-routing
        """

        def db_for_read(self, model: type[Model], **hints: Any) -> str | None: ...

        def db_for_write(self, model: type[Model], **hints: Any) -> str | None: ...

        def allow_relation(self, obj1: type[Model], obj2: type[Model], **hints: Any) -> bool | None: ...

        def allow_migrate(
            self, db: str, app_label: str, model_name: str | None = None, **hints: Any
        ) -> bool | None: ...

else:
    TypedDatabaseRouter = object



from typing import TYPE_CHECKING

# Re-export stubs-only classes RelatedManger and ManyRelatedManager.
# These are fake, Django defines these inside function body.
if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from django.db.models.fields.related_descriptors import ManyRelatedManager as ManyRelatedManager

    # noinspection PyUnresolvedReferences
    from django.db.models.fields.related_descriptors import RelatedManager as RelatedManager

else:
    from typing import Protocol, TypeVar

    _T = TypeVar("_T")
    _Through = TypeVar("_Through")

    # Define as `Protocol` to prevent them being used with `isinstance()`.
    # These actually inherit from `BaseManager`.
    class RelatedManager(Protocol[_T]):
        pass

    class ManyRelatedManager(Protocol[_T, _Through]):
        pass



from collections.abc import Mapping
from typing import Annotated, Any, Generic, TypeVar

from django.db.models.base import Model

# Really, we would like to use TypedDict as a bound, but it's not possible
_Annotations = TypeVar("_Annotations", covariant=True, bound=Mapping[str, Any])


class Annotations(Generic[_Annotations]):
    """Use as `Annotations[MyTypedDict]`"""

    pass


_T = TypeVar("_T", bound=Model)

WithAnnotations = Annotated[_T, Annotations[_Annotations]]
"""Alias to make it easy to annotate the model `_T` as having annotations
`_Annotations` (a `TypedDict`).

Use as `WithAnnotations[MyModel, MyTypedDict]`.
"""



from typing import Any, Protocol


# Used internally by mypy_django_plugin.
class AnyAttrAllowed(Protocol):
    def __getattr__(self, item: str) -> Any: ...

    def __setattr__(self, item: str, value: Any) -> None: ...



from .aliases import QuerySetAny as QuerySetAny
from .aliases import StrOrPromise, StrPromise
from .aliases import ValuesQuerySet as ValuesQuerySet
from .annotations import Annotations as Annotations
from .annotations import WithAnnotations as WithAnnotations
from .patch import monkeypatch as monkeypatch
from .types import AnyAttrAllowed as AnyAttrAllowed

__all__ = [
    "monkeypatch",
    "QuerySetAny",
    "ValuesQuerySet",
    "WithAnnotations",
    "Annotations",
    "AnyAttrAllowed",
    "StrPromise",
    "StrOrPromise",
]



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import ClassVar, Literal

    from django.db.models import BaseConstraint, Index, OrderBy
    from django.utils.datastructures import _ListOrTuple

    from django_stubs_ext import StrOrPromise

    class TypedModelMeta:
        """
        Typed base class for Django Model `class Meta:` inner class. At runtime this is just an alias to `object`.

        Most attributes are the same as `django.db.models.options.Options`. Options has some additional attributes and
        some values are normalized by Django.
        """

        abstract: ClassVar[bool]  # default: False
        app_label: ClassVar[str]
        base_manager_name: ClassVar[str]
        db_table: ClassVar[str]
        db_table_comment: ClassVar[str]
        db_tablespace: ClassVar[str]
        default_manager_name: ClassVar[str]
        default_related_name: ClassVar[str]
        get_latest_by: ClassVar[str | Sequence[str]]
        managed: ClassVar[bool]  # default: True
        order_with_respect_to: ClassVar[str]
        ordering: ClassVar[Sequence[str | OrderBy]]
        permissions: ClassVar[_ListOrTuple[tuple[str, StrOrPromise]]]
        default_permissions: ClassVar[Sequence[str]]  # default: ("add", "change", "delete", "view")
        proxy: ClassVar[bool]  # default: False
        required_db_features: ClassVar[_ListOrTuple[str]]
        required_db_vendor: ClassVar[Literal["sqlite", "postgresql", "mysql", "oracle"]]
        select_on_save: ClassVar[bool]  # default: False
        indexes: ClassVar[_ListOrTuple[Index]]
        unique_together: ClassVar[Sequence[Sequence[str]]]
        index_together: ClassVar[Sequence[Sequence[str]]]  # Deprecated in Django 4.2
        constraints: ClassVar[_ListOrTuple[BaseConstraint]]
        verbose_name: ClassVar[StrOrPromise]
        verbose_name_plural: ClassVar[StrOrPromise]

else:
    TypedModelMeta = object
